from collections import namedtuple

import numpy as np

from jesse.helpers import slice_candles
from jesse.indicators.ma import ma

Stochastic = namedtuple('Stochastic', ['k', 'd'])


def stoch(candles: np.ndarray, fastk_period: int = 14, slowk_period: int = 3, slowk_matype: int = 0,
          slowd_period: int = 3, slowd_matype: int = 0, sequential: bool = False) -> Stochastic:
    """
    The Stochastic Oscillator

    :param candles: np.ndarray
    :param fastk_period: int - default: 14
    :param slowk_period: int - default: 3
    :param slowk_matype: int - default: 0
    :param slowd_period: int - default: 3
    :param slowd_matype: int - default: 0
    :param sequential: bool - default: False

    :return: Stochastic(k, d)
    """
    if any(matype in (24, 29) for matype in (slowk_matype, slowd_matype)):
        raise ValueError("VWMA (matype 24) and VWAP (matype 29) cannot be used in stochastic indicator.")

    candles = slice_candles(candles, sequential)

    candles_close = candles[:, 2]
    candles_high = candles[:, 3]
    candles_low = candles[:, 4]
    
    hh = _rolling_max(candles_high, fastk_period)
    ll = _rolling_min(candles_low, fastk_period)

    stoch_val = 100 * (candles_close - ll) / (hh - ll)
    k = ma(stoch_val, period=slowk_period, matype=slowk_matype, sequential=True)
    d = ma(k, period=slowd_period, matype=slowd_matype, sequential=True)

    if sequential:
        return Stochastic(k, d)
    else:
        return Stochastic(k[-1], d[-1])

def _rolling_max(x, window):
    if len(x) < window:
        return np.full(x.shape, np.nan, dtype=np.float64)
    windows = np.lib.stride_tricks.sliding_window_view(x, window_shape=window)
    result = np.full(x.shape, np.nan, dtype=np.float64)
    result[window - 1:] = np.max(windows, axis=1)
    return result

def _rolling_min(x, window):
    if len(x) < window:
        return np.full(x.shape, np.nan, dtype=np.float64)
    windows = np.lib.stride_tricks.sliding_window_view(x, window_shape=window)
    result = np.full(x.shape, np.nan, dtype=np.float64)
    result[window - 1:] = np.min(windows, axis=1)
    return result



from collections import namedtuple

import numpy as np

from jesse.helpers import slice_candles

DI = namedtuple('DI', ['plus', 'minus'])


def di(candles: np.ndarray, period: int = 14, sequential: bool = False) -> DI:
    """
    DI - Directional Indicator

    :param candles: np.ndarray, where columns are expected to be: index 2: close, index 3: high, index 4: low.
    :param period: int - default: 14
    :param sequential: bool - default: False

    :return: DI(plus, minus)
    """
    candles = slice_candles(candles, sequential)
    n = len(candles)
    if n < 2:
        if sequential:
            return DI(np.full(n, np.nan), np.full(n, np.nan))
        else:
            return DI(np.nan, np.nan)

    # Extract high, low, and close assuming columns: index 3 -> high, index 4 -> low, index 2 -> close
    high = candles[:, 3]
    low = candles[:, 4]
    close = candles[:, 2]

    # Calculate directional movements
    up_move = high[1:] - high[:-1]
    down_move = low[:-1] - low[1:]
    plus_dm = np.where((up_move > down_move) & (up_move > 0), up_move, 0)
    minus_dm = np.where((down_move > up_move) & (down_move > 0), down_move, 0)

    # Calculate True Range (TR)
    high_low = high[1:] - low[1:]
    high_close = np.abs(high[1:] - close[:-1])
    low_close = np.abs(low[1:] - close[:-1])
    tr = np.maximum.reduce([high_low, high_close, low_close])

    m = len(tr)  # m = n - 1
    # Initialize arrays for Wilder's smoothing
    atr = np.full(m, np.nan)
    plus_smoothed = np.full(m, np.nan)
    minus_smoothed = np.full(m, np.nan)

    if m < period:
        plus_DI_arr = np.full(n, np.nan)
        minus_DI_arr = np.full(n, np.nan)
        if sequential:
            return DI(plus_DI_arr, minus_DI_arr)
        else:
            return DI(np.nan, np.nan)

    # Initial Wilder's smoothing
    atr[period - 1] = np.sum(tr[:period]) / period
    plus_smoothed[period - 1] = np.sum(plus_dm[:period])
    minus_smoothed[period - 1] = np.sum(minus_dm[:period])

    # Wilder's smoothing recursion
    for i in range(period, m):
        atr[i] = (atr[i - 1] * (period - 1) + tr[i]) / period
        plus_smoothed[i] = (plus_smoothed[i - 1] * (period - 1) + plus_dm[i]) / period
        minus_smoothed[i] = (minus_smoothed[i - 1] * (period - 1) + minus_dm[i]) / period

    # Prepare DI arrays: first 'period' candles are not computed (set to NaN)
    plus_DI_arr = np.full(n, np.nan)
    minus_DI_arr = np.full(n, np.nan)
    valid_indices = np.arange(period, n)
    plus_DI_arr[valid_indices] = np.where(atr[valid_indices - 1] == 0, 0, 100 * plus_smoothed[valid_indices - 1] / atr[valid_indices - 1])
    minus_DI_arr[valid_indices] = np.where(atr[valid_indices - 1] == 0, 0, 100 * minus_smoothed[valid_indices - 1] / atr[valid_indices - 1])

    if sequential:
        return DI(plus_DI_arr, minus_DI_arr)
    else:
        return DI(plus_DI_arr[-1], minus_DI_arr[-1])



from typing import Union

import numpy as np
from numba import njit

from jesse.helpers import get_candle_source, slice_candles


@njit(cache=True)
def _ema(source: np.ndarray, period: int) -> np.ndarray:
    """
    Compute the Exponential Moving Average using a loop.
    """
    n = len(source)
    result = np.full(n, np.nan)
    if n < period:
        return result
    alpha = 2 / (period + 1)
    # Initialize EMA with the simple average of the first "period" values
    initial = np.mean(source[:period])
    result[period - 1] = initial
    prev = initial
    for i in range(period, n):
        current = alpha * source[i] + (1 - alpha) * prev
        result[i] = current
        prev = current
    return result

def ema(candles: np.ndarray, period: int = 5, source_type: str = "close", sequential: bool = False) -> Union[
    float, np.ndarray]:
    """
    EMA - Exponential Moving Average using Numba for optimization

    :param candles: np.ndarray
    :param period: int - default: 5
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    if len(candles.shape) == 1:
        source = candles
    else:
        candles = slice_candles(candles, sequential)
        source = get_candle_source(candles, source_type=source_type)

    result = _ema(source, period)
    return result if sequential else result[-1]



from collections import namedtuple

import numpy as np
from numba import njit

from jesse.helpers import get_candle_source, same_length, slice_candles

StochasticRSI = namedtuple('StochasticRSI', ['k', 'd'])


@njit(cache=True)
def _rolling_window(a, window):
    """Helper function to create rolling windows of data"""
    shape = a.shape[:-1] + (a.shape[-1] - window + 1, window)
    strides = a.strides + (a.strides[-1],)
    return np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)


@njit(cache=True)
def _calculate_stoch(data, period_stoch, k_period, d_period):
    """Calculate stochastic oscillator values"""
    # Calculate rolling min and max
    rolling_mins = np.array([np.min(w) for w in _rolling_window(data, period_stoch)])
    rolling_maxs = np.array([np.max(w) for w in _rolling_window(data, period_stoch)])

    # Calculate %K
    k_fast = 100 * (data[period_stoch-1:] - rolling_mins) / (rolling_maxs - rolling_mins)

    # Calculate smoothed %K (which becomes the final %K)
    k = np.zeros_like(k_fast)
    for i in range(k_period-1, len(k_fast)):
        k[i] = np.mean(k_fast[i-k_period+1:i+1])

    # Calculate %D (SMA of %K)
    d = np.zeros_like(k)
    for i in range(d_period-1, len(k)):
        d[i] = np.mean(k[i-d_period+1:i+1])

    return k, d


@njit(cache=True)
def _calculate_rsi(source, period):
    n = len(source)
    rsi = np.full(n, np.nan)  # initialize all values as NaN
    if n <= period:
        return rsi

    # Vectorized computation of price differences
    delta = source[1:] - source[:-1]

    # Compute ups and downs using numpy where
    ups = np.where(delta > 0, delta, 0.0)
    downs = np.where(delta < 0, -delta, 0.0)

    # Initial average gain and loss
    avg_gain = np.mean(ups[:period])
    avg_loss = np.mean(downs[:period])

    # First RSI value at index 'period'
    if avg_loss == 0:
        rsi[period] = 100.0
    else:
        rsi[period] = 100.0 - 100.0 / (1.0 + avg_gain / avg_loss)

    # Recursive calculation for subsequent RSI values
    for i in range(period, len(ups)):
        avg_gain = (avg_gain * (period - 1) + ups[i]) / period
        avg_loss = (avg_loss * (period - 1) + downs[i]) / period
        if avg_loss == 0:
            rsi[i + 1] = 100.0
        else:
            rsi[i + 1] = 100.0 - 100.0 / (1.0 + avg_gain / avg_loss)
    
    return rsi


def srsi(candles: np.ndarray, period: int = 14, period_stoch: int = 14, k: int = 3, d: int = 3,
         source_type: str = "close", sequential: bool = False) -> StochasticRSI:
    """
    Stochastic RSI

    :param candles: np.ndarray
    :param period: int - default: 14 - RSI Length
    :param period_stoch: int - default: 14 - Stochastic Length
    :param k: int - default: 3
    :param d: int - default: 3
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: StochasticRSI(k, d)
    """
    candles = slice_candles(candles, sequential)

    source = get_candle_source(candles, source_type=source_type)
    rsi_np = _calculate_rsi(source, period)
    rsi_np = rsi_np[np.logical_not(np.isnan(rsi_np))]

    # Calculate stochastic values
    fast_k, fast_d = _calculate_stoch(rsi_np, period_stoch, k, d)

    if sequential:
        # Pad the beginning with NaN to match the original length
        pad_length = len(candles) - len(fast_k)
        fast_k = np.concatenate((np.full(pad_length, np.nan), fast_k))
        fast_d = np.concatenate((np.full(pad_length, np.nan), fast_d))
        return StochasticRSI(fast_k, fast_d)
    else:
        return StochasticRSI(fast_k[-1], fast_d[-1])



from typing import Union

import numpy as np
from numba import njit

from jesse.helpers import get_candle_source, same_length, slice_candles


@njit
def _zlema_fast(source: np.ndarray, period: int) -> np.ndarray:
    lag = (period - 1) / 2
    lag = int(lag)

    # Pre-allocate the output array
    res = np.zeros_like(source)

    # Calculate the smoothing factor
    alpha = 2 / (period + 1)

    # Calculate ema_data = price + (price - price_lag)
    ema_data = np.zeros_like(source)
    for i in range(lag, len(source)):
        ema_data[i] = source[i] + (source[i] - source[i - lag])

    # First value is a simple copy
    res[lag] = ema_data[lag]

    # Calculate ZLEMA
    for i in range(lag + 1, len(source)):
        res[i] = (alpha * ema_data[i]) + ((1 - alpha) * res[i - 1])

    return res


def zlema(candles: np.ndarray, period: int = 20, source_type: str = "close", sequential: bool = False) -> Union[
        float, np.ndarray]:
    """
    Zero-Lag Exponential Moving Average

    :param candles: np.ndarray
    :param period: int - default: 20
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    if len(candles.shape) == 1:
        source = candles
    else:
        candles = slice_candles(candles, sequential)
        source = get_candle_source(candles, source_type=source_type)

    res = _zlema_fast(source, period)

    return same_length(candles, res) if sequential else res[-1]
